package net.eq2online.macros.scripting;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.regex.Pattern;

import net.eq2online.console.Log;
import net.eq2online.macros.compatibility.Reflection;
import net.eq2online.macros.scripting.api.IScriptAction;

/**
 * Base class for script actions, manages registration of script actions
 * 
 * @author Adam Mummery-Smith
 */
public abstract class ScriptActionBase implements IScriptAction
{
	/**
	 * Name of this action
	 */
	protected String actionName;

	/**
	 * Mapping of action names to action providers
	 */
	protected static HashMap<String, IScriptAction> actions = new HashMap<String, IScriptAction>();
	
	/**
	 * Regex used to match actions when highlighting, each registered action is added to this regex so that
	 * it can be matched when required by the GUI. Currently only used by the text editor. 
	 */
	protected static Pattern actionRegex = Pattern.compile("");
	
	/**
	 * Protected ctor, used for child instance singletons
	 * 
	 * @param actionName
	 */
	protected ScriptActionBase(String actionName)
	{
		this.actionName = actionName.toLowerCase(); 
	}
	
	/* (non-Javadoc)
	 * @see net.eq2online.macros.scripting.api.IScriptAction#getName()
	 */
	@Override
	public String getName()
	{
		return this.actionName;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString()
	{
		return this.actionName;
	}

	/**
	 * Registration callback for script core
	 * 
	 * @param newAction
	 */
	protected static boolean registerAction(IScriptAction newAction)
	{
		if (actions.containsKey(newAction.toString()))
			return false;
		
		actions.put(newAction.toString(), newAction);
		ScriptCore.getDocumentor().setDocumentation(newAction);
		ScriptActionBase.updateScriptActionRegex();
		
		return true;
	}

	/**
	 * Updates the regex used to recognise script actions in the text editor
	 */
	protected static void updateScriptActionRegex()
	{
		String actionList = "";
		char separator = '(';
		
		for (IScriptAction action : actions.values())
		{
			actionList += separator + action.toString();
			separator = '|';
		}
		
		actionList += ")";
		
		actionRegex = Pattern.compile(actionList + "([\\(;]|$)", Pattern.CASE_INSENSITIVE);
	}

	/**
	 * Initialise the script engine with available script actions by reflection
	 * 
	 * @param logger
	 */
	@SuppressWarnings("unchecked")
	public static void initActions(IErrorLogger logger)
	{
		try
		{
			Class<? extends ScriptActionBase> packageClass = (Class<? extends ScriptActionBase>)Class.forName("net.eq2online.macros.scripting.actions.lang.ScriptActionAssign");
			LinkedList<Class<?>> actions = Reflection.getSubclassesFor(ScriptActionBase.class, packageClass, "ScriptAction", logger);
			int loadedActions = 0;
			
			for (Class<?> action : actions)
			{
				try
				{
					IScriptAction newAction = (IScriptAction)action.newInstance();
					if (newAction != null)
					{
						if (ScriptActionBase.registerAction(newAction))
							loadedActions++;
					}
				}
				catch (Exception ex) {}
			}

			Log.info("Script engine initialised, registered {0} script action(s)", loadedActions);
			//ScriptCore.getDocumentor().writeXml(new File(MacroModCore.getMacrosDirectory(), "scriptactions.xml"));
			
			if (loadedActions == 0)
			{
				logger.logError("Script engine initialisation error");
			}
		}
		catch (ClassNotFoundException ex)
		{
			logger.logError("Script engine initialisation error, package not found");
		}
	}
}