package net.eq2online.macros.scripting;

import java.io.File;
import java.io.FileInputStream;
import java.io.FilenameFilter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import net.eq2online.console.Log;
import net.eq2online.macros.scripting.api.APIVersion;
import net.eq2online.macros.scripting.api.IMacro;
import net.eq2online.macros.scripting.api.IMacroEventManager;
import net.eq2online.macros.scripting.api.IMacroEventProvider;
import net.eq2online.macros.scripting.api.IMacrosAPIModule;
import net.eq2online.macros.scripting.api.IScriptAction;
import net.eq2online.macros.scripting.api.IScriptActionProvider;
import net.eq2online.macros.scripting.api.IScriptParser;
import net.eq2online.macros.scripting.api.IScriptedIterator;
import net.eq2online.macros.scripting.api.IVariableListener;
import net.eq2online.macros.scripting.api.IVariableProvider;
import net.minecraft.launchwrapper.LaunchClassLoader;

/**
 * Manages loading and registration of script modules and acts as a central repository for some
 * script engine resources.
 * 
 * @author Adam Mummery-Smith
 */
public abstract class ScriptCore
{
	/**
	 * Active Script Action Provider module
	 */
	private static IScriptActionProvider activeProvider;
	
	/**
	 * Active event manager
	 */
	private static IMacroEventManager eventManager;
	
	/**
	 * Default script parser
	 */
	private static IScriptParser defaultParser;
	
	/**
	 * Directory containing script addon modules
	 */
	private static File modulesDir;
	
	/**
	 * True if the script module initialisation was already performed
	 */
	private static boolean initDone;
	
	/**
	 * Iterators
	 */
	private static HashMap<String, Class<? extends IScriptedIterator>> iterators = new HashMap<String, Class<? extends IScriptedIterator>>();
	
	/**
	 * Documentation provider
	 */
	private static IDocumentor documentor;
	
	/**
	 * Initialise the script core and load the addon modules 
	 * 
	 * @param macrosPath
	 * @param provider
	 */
	public static void init(File macrosPath, IScriptActionProvider provider, IMacroEventManager eventManager, IScriptParser defaultParser, IErrorLogger logger, IDocumentor documentor)
	{
		if (!ScriptCore.initDone && macrosPath != null && provider != null)
		{
			ScriptCore.initDone = true;
			ScriptCore.modulesDir = new File(macrosPath, "/modules");
			
			try
			{
				ScriptCore.modulesDir.mkdirs();
			}
			catch (Exception ex) {}
			
			ScriptCore.registerScriptActionProvider(provider);
			ScriptCore.eventManager = eventManager;
			ScriptCore.defaultParser = defaultParser;
			ScriptCore.documentor = documentor;
			ScriptCore.loadModules(logger);
		}
	}

	/**
	 * Register a new script action with the script core
	 * 
	 * @param action
	 */
	public static void registerScriptAction(IScriptAction action)
	{
		try
		{
			ScriptActionBase.registerAction(action);
		}
		catch (Throwable ex)
		{
			ex.printStackTrace();
		}
	}
	
	/**
	 * Register a script action provider with the core
	 * 
	 * @param provider
	 */
	public static void registerScriptActionProvider(IScriptActionProvider provider)
	{
		ScriptCore.activeProvider = provider;
	}
	
	/**
	 * Get the currently registered script action provider
	 * 
	 * @return
	 */
	public static IScriptActionProvider getScriptActionProvider()
	{
		return ScriptCore.activeProvider;
	}

	/**
	 * Gets the default script parser
	 * 
	 * @return
	 */
	public static IScriptParser getDefaultParser()
	{
		return ScriptCore.defaultParser;
	}

	/**
	 * Get the currently bound documentation provider
	 * 
	 * @return
	 */
	public static IDocumentor getDocumentor()
	{
		return ScriptCore.documentor;
	}
	
	/**
	 * Register a new variable provider with the active script action provider
	 * 
	 * @param variableProvider
	 */
	public static void registerVariableProvider(IVariableProvider variableProvider)
	{
		if (ScriptCore.activeProvider != null)
		{
			ScriptCore.activeProvider.registerVariableProvider(variableProvider);
		}
	}
	
	/**
	 * Register a new variable listener with the active script action provider
	 * 
	 * @param variableListener
	 */
	public static void registerVariableListener(IVariableListener variableListener)
	{
		if (ScriptCore.activeProvider != null)
		{
			ScriptCore.activeProvider.registerVariableListener(variableListener);
		}
	}
	
	/**
	 * Set a variable value
	 * 
	 * @param provider
	 * @param macro
	 * @param variableName
	 * @param variableValue
	 */
	public static void setVariable(IScriptActionProvider provider, IMacro macro, String variableName, String variableValue)
	{
		int intValue = ScriptCore.tryParseInt(variableValue, 0);
		boolean boolValue = ScriptCore.parseBoolean(variableValue, intValue);
		provider.setVariable(macro, variableName, variableValue, intValue, boolValue);
	}
	
	/**
	 * @param provider
	 * @param macro
	 * @param variableName
	 * @param variableValue
	 */
	public static void setVariable(IScriptActionProvider provider, IMacro macro, String variableName, int variableValue)
	{
		if (variableName.length() > 0)
		{
			provider.setVariable(macro, variableName, String.valueOf(variableValue), variableValue, variableValue != 0);
		}
	}
	
	/**
	 * Parse variables in the supplied string using the supplied context objects
	 * 
	 * @param provider
	 * @param macro
	 * @param text
	 * @param quoteStrings
	 * @return
	 */
	public static String parseVars(IScriptActionProvider provider, IMacro macro, String text, boolean quoteStrings)
	{
		return new VariableExpander(provider, macro, text, quoteStrings).toString();
	}

	/**
	 * Register a new iterator
	 * 
	 * @param iteratorName
	 * @param iterator
	 */
	public static void registerIterator(String iteratorName, Class<? extends IScriptedIterator> iterator)
	{
		if (ScriptCore.iterators.containsKey(iteratorName))
		{
			return;
//			throw new IllegalArgumentException("Attempted to register a duplicate iterator for key '" + iteratorName + "'");
		}
		
		ScriptCore.iterators.put(iteratorName, iterator);
	}
	
	/**
	 * Get an iterator by key
	 * 
	 * @param iteratorName
	 * @return
	 */
	public static Class<? extends IScriptedIterator> getIterator(String iteratorName)
	{
		return ScriptCore.iterators.get(iteratorName);
	}
	
	public static void registerEventProvider(IMacroEventProvider eventProvider)
	{
		if (ScriptCore.eventManager != null)
		{
			ScriptCore.eventManager.registerEventProvider(eventProvider);
		}
	}
	
	/**
	 * Load the script modules
	 */
	protected static void loadModules(IErrorLogger logger)
	{
		if (ScriptCore.modulesDir.exists() && ScriptCore.modulesDir.isDirectory())
		{
			try
			{
				LaunchClassLoader classLoader = (LaunchClassLoader)ScriptCore.class.getClassLoader();

				// Enumerate module files
				ArrayList<File> moduleFiles = new ArrayList<File>();
				
				for (File file : modulesDir.listFiles(new FilenameFilter() {
					@Override
					public boolean accept(File dir, String name) { return name.startsWith("module_") && (name.endsWith(".zip") || name.endsWith(".jar")); }
					}
				))
				{
					moduleFiles.add(file);
				}
				
				// Add modules files to class path
				for (File module : moduleFiles)
				{
					if (module.isFile())
					{
						classLoader.addURL(module.toURI().toURL());
					}
				}
				
				// Load actions and providers from modules
				for (File module : moduleFiles)
				{
					LoadedModuleInfo lmi = new LoadedModuleInfo(module);
					
					if (module.isFile() && module.getName().startsWith("module_"));
					{
						ZipInputStream zipinputstream = new ZipInputStream(new FileInputStream(module));
						
						do
						{
							ZipEntry zipentry = zipinputstream.getNextEntry();
							if (zipentry == null) break;
							
							String className = new File(zipentry.getName()).getName();
							
							if (!zipentry.isDirectory() && className.endsWith(".class") && !className.contains("$"))
							{
								String fullClassName = zipentry.getName().split("\\.")[0];
								IMacrosAPIModule newModule = null;
								
								if (className.startsWith("ScriptAction"))
								{
									newModule = lmi.addAction(ScriptCore.addModule(classLoader, IScriptAction.class, "action", fullClassName, logger));
								}
								else if (className.startsWith("VariableProvider"))
								{
									newModule = lmi.addProvider(ScriptCore.addModule(classLoader, IVariableProvider.class, "variable provider", fullClassName, logger));
								}
								else if (className.startsWith("ScriptedIterator"))
								{
									newModule = lmi.addIterator(ScriptCore.addModule(classLoader, IScriptedIterator.class, "iterator", fullClassName, logger));
								}
								else if (className.startsWith("EventProvider"))
								{
									newModule = lmi.addEventProvider(ScriptCore.addModule(classLoader, IMacroEventProvider.class, "event provider", fullClassName, logger));
								}
								else
								{
									continue;
								}
								
								if (newModule == null)
								{
									if (logger != null) logger.logError("API: Error initialising " + module.getName());
									break;
								}
							}
						}
						while (true);
						
						zipinputstream.close();
					}
					
					lmi.printStatus();
				}
			}
			catch (Throwable th)
			{
			}
		}
	}

	/**
	 * Add a new action class from a module
	 * 
	 * @param classLoader
	 * @param className
	 */
	private static <ModuleType extends IMacrosAPIModule> ModuleType addModule(ClassLoader classLoader, Class<ModuleType> moduleClassType, String moduleType, String className, IErrorLogger logger)
	{
		try
		{
			className = className.replace('/', '.');
			Class<?> moduleClass = classLoader.loadClass(className);
			
			if (moduleClassType.isAssignableFrom(moduleClass))
			{
				if (!ScriptCore.checkAPIVersion(moduleClass))
				{
					Log.info("Macros: API Error. Not loading custom {0} in {1}, bad API version.", moduleType, className);
					if (logger != null) logger.logError("API: Not loading " + moduleClass.getSimpleName() + ", bad API version");
					return null;
				}
				
				@SuppressWarnings("unchecked")
				ModuleType customScriptAction = (ModuleType)moduleClass.newInstance();
				
				if (customScriptAction != null)
				{
					customScriptAction.onInit();
					return customScriptAction;
				}
			}
		}
		catch (Throwable ex)
		{
			ex.printStackTrace();
		}
		
		return null;
	}
	
	/**
	 * If you can't work out what this function does then there's something wrong with your head
	 * 
	 * @param moduleClass
	 * @return
	 */
	private static final boolean checkAPIVersion(Class<?> moduleClass)
	{
		APIVersion versionAnnotation = moduleClass.getAnnotation(APIVersion.class);
		return (versionAnnotation != null && versionAnnotation.value() == 13); // TODO API Version
	}
	
	/**
	 * @param variableValue
	 * @param intValue
	 * @return
	 */
	public static boolean parseBoolean(String variableValue)
	{
		int intValue = ScriptCore.tryParseInt(variableValue, 0);
		return ScriptCore.parseBoolean(variableValue, intValue);
	}

	/**
	 * @param variableValue
	 * @param intValue
	 * @return
	 */
	public static boolean parseBoolean(String variableValue, int intValue)
	{
		return variableValue == null || variableValue.toLowerCase().equals("true") || intValue != 0;
	}

	/**
	 * Safe function to parse an integer, returns 0 on failure
	 * 
	 * @param value String parameter to parse
	 * @return Parsed value or 0 if parsing failed
	 */
	public static int tryParseInt(String value, int defaultValue)
	{
		if (value == null) return defaultValue;
		try { return Integer.parseInt(value.trim().replaceAll(",", "")); } catch (NumberFormatException ex) {} return defaultValue;
	}
	
	/**
	 * Safe function to parse a long, returns 0 on failure
	 * 
	 * @param value String parameter to parse
	 * @return Parsed value or 0 if parsing failed
	 */
	public static long tryParseLong(String value, long defaultValue)
	{
		try { return Long.parseLong(value.trim()); } catch (NumberFormatException ex) {} return defaultValue;
	}
	
	/**
	 * Safe function to parse a float, returns 0 on failure
	 * 
	 * @param value String parameter to parse
	 * @return Parsed value or 0 if parsing failed
	 */
	public static float tryParseFloat(String value, float defaultValue)
	{
		try { return Float.parseFloat(value.trim()); } catch (NumberFormatException ex) {} return defaultValue;
	}
	
	public static int tryParseIntOffset(String value, int defaultValue, int source)
	{
		int offsetDirection = 1;
		
		if (value.startsWith("+"))
		{
			value = value.substring(1);
		}
		else if (value.startsWith("-"))
		{
			offsetDirection = -1;
			value = value.substring(1);
		}
		else
		{
			source = 0;
		}
		
		return source + (ScriptCore.tryParseInt(value, defaultValue) * offsetDirection);
	}
}
